from .dynamics import CalciumToSpike
from .dynamics import SpikeToCalcium
from .dynamics import get_dynamics
from .dynamics import get_rdynamics
from .evaluate import evaluate
from .evaluate import fix_kind

__all__ = [
    'CalciumToSpike',
    'SpikeToCalcium',
    'get_dynamics',
    'get_rdynamics',
    'fix_kind',
    'evaluate',
]



from collections import namedtuple

from .regularizer import L2
from .regularizer import MaxNormNonNegativeL1
from .regularizer import NonNegativeL1
from .regularizer import Regularizer

Penalty = namedtuple('Penalty', 'la lu lb bs bt')


def get_penalty(penalty, stage=None):
    if isinstance(penalty, Penalty):
        return penalty
    out = []
    for p in (penalty.la, penalty.lu):
        match p:
            case {'type': 'NoPenalty'}:
                out.append((Regularizer(), ()))
            case {'type': 'L2', 'factor': fac}:
                out.append((L2, (fac,)))
            case {'type': 'NonNegativeL1', 'factor': fac}:
                out.append((NonNegativeL1(), (fac,)))
            case {'type': 'MaxNormNonNegativeL1', 'factor': fac}:
                out.append((MaxNormNonNegativeL1(), (fac,)))
            case {'type': 'MaxNormNonNegativeL1', 'annealing': fac, 'max': maxval}:
                out.append((MaxNormNonNegativeL1(), (min(stage * fac, maxval),)))
            case _:
                raise ValueError()
    out += [penalty.lb, penalty.bs, penalty.bt]
    return Penalty(*out)



import sys
from datetime import datetime
from importlib import import_module
from multiprocessing import Process
from pathlib import Path

import numpy as np

from ..plot import seg_max_fig
from ..plot import spike_image
from ..plot import to_csv
from ..plot import to_multipage_tif
from .common import finish
from .common import print_stats
from .common import set_env


def call(name, *args, **kwargs):
    def wrap(cfg, *args, **kwargs):
        set_env(cfg)
        return target(cfg, *args, **kwargs)

    target = getattr(import_module(f'hotaru.cui.{name}'), name)
    p = Process(target=wrap, args=args, kwargs=kwargs)
    p.start()
    p.join()
    if p.exitcode != 0:
        sys.exit()


def run(cfg):
    now = datetime.now().isoformat()

    for stage in range(cfg.max_train_step + 1):
        if stage == 0:
            call('normalize', cfg)
            call('init', cfg)
        else:
            call('spatial', cfg, stage)
        call('temporal', cfg, stage)

        print_stats(cfg, stage)
        if finish(cfg, stage):
            break

    path = Path(cfg.outputs.outputs.dir) / now
    path.mkdir(parents=True, exist_ok=True)
    to_csv(cfg, stage, path)
    to_multipage_tif(cfg, stage, path)
    seg_max_fig(cfg, stage).write_image(path / 'run_footprints.pdf')
    spike_image(cfg, stage, tsel=np.arange(4096))[0].save(path / 'run_spike.pdf')



from .clean import clean
from .clean import clean_footprints
from .find import find_peaks
from .make import make_footprints
from .radius import get_radius
from .reduce import reduce_peaks
from .reduce import reduce_peaks_simple

__all__ = [
    'get_radius',
    'find_peaks',
    'reduce_peaks',
    'reduce_peaks_simple',
    'make_footprints',
    'clean',
    'clean_footprints',
]



import jax.numpy as jnp


def get_radius(cfg):
    if isinstance(cfg, tuple):
        return cfg
    match cfg['type']:
        case 'logscale':
            r = jnp.geomspace(cfg['min'], cfg['max'], cfg['num'])
        case 'linear':
            r = jnp.linspace(cfg['min'], cfg['max'], cfg['num'])
        case 'list':
            r = jnp.array(cfg['val'])
    return tuple(float(ri) for ri in r)



import numpy as np
import pandas as pd
import plotly.express as px
import plotly.io as pio

pio.kaleido.scope.mathjax = None


def plot_stats(paths, imgs, labels, dlabels, **kwargs):
    dfs = []
    for dlabel, path in zip(dlabels, paths, strict=False):
        df = pd.DataFrame()
        for label, img in zip(labels, imgs, strict=False):
            img = np.load(f'{path}/{img}.npy')
            df[label] = img.ravel()
        df['Data'] = dlabel
        dfs.append(df)
    df = pd.concat(dfs, axis=0)
    fig = px.scatter(df, x=labels[1], y=labels[2], color=labels[0], facet_col='Data', **kwargs)
    fig.update_xaxes(
        title_font_size=11,
    )
    fig.update_yaxes(
        title_font_size=11,
    )
    fig.update_layout(
        coloraxis_colorbar_title_font_size=11,
        template='none',
        font_size=11,
        margin={'l': 50, 'r': 20, 't': 20, 'b': 50},
        annotations=[
            {'font': {'size': 11}, 'text': dlabels[i]} for i in range(len(fig.layout.annotations))
        ],
    )
    return fig



import numpy as np
import plotly.express as px

from ..cui.common import all_stats


def cell_num_fig(cfg, **kwargs):
    kwargs.setdefault('template', 'none')
    kwargs.setdefault('margin', {'l': 40, 'r': 10, 't': 20, 'b': 35})

    num = [np.count_nonzero(s.kind == 'cell') for s in all_stats(cfg)]

    fig = px.line(x=np.arange(len(num)), y=num)
    fig.add_annotation(x=1, y=num[1], ax=2, text='A')
    fig.add_annotation(x=len(num) - 1, y=num[-1], text='B')
    fig.update_xaxes(
        title_text='epoch',
        title_font_size=11,
    )
    fig.update_yaxes(
        title_text='Number of cells',
        title_font_size=11,
    )
    fig.update_layout(**kwargs)
    return fig



import pandas as pd
import plotly.express as px
import plotly.io as pio

from ..cui.common import load

pio.kaleido.scope.mathjax = None


def dense_sig_fig(cfg, stages, thr_sig=0, label='', thr_udense=1.0):
    num = []
    dfs = []
    for stage in stages:
        df, _ = load(cfg, 'evaluate', stage)
        df = df.query("kind != 'remove'")
        df.loc[df.udense > thr_udense, 'kind'] = 'background'
        df['col'] = stage
        df['size'] = 1
        df['symbol'] = 1
        dfs.append(df)
        num.append(df.query("kind == 'cell'").shape[0])
    df = pd.concat(dfs, axis=0)
    fig = px.scatter(
        df,
        x='udense',
        y='signal',
        color='kind',
        size='size',
        size_max=3,
        opacity=0.5,
        facet_col='col',
    )
    for i in range(len(fig.data)):
        fig.data[i].marker.line.width = 0

    for n, stage in zip(num, stages, strict=False):
        fig.update_annotations(
            selector={'text': f'col={stage}'}, y=0.8, text=f'{label}{stage}; num={n}'
        )
    fig.update_layout(
        template='none',
        font_size=11,
        showlegend=False,
        width=600,
        height=150,
        margin={'l': 35, 'r': 10, 't': 20, 'b': 32},
    )
    return fig


def dense_sig_multi_fig(paths, stages, thr_sig=0, thr_udense=1.0):
    num = []
    dfs = []
    for i, (name, path) in enumerate(paths.items()):
        for j, stage in enumerate(stages[i]):
            if stage == 0:
                pathx = f'{path}/../../{stage:03d}stats.csv'
            else:
                pathx = f'{path}/{stage:03d}stats.csv'
            df = pd.read_csv(pathx)
            df = df.query("kind != 'remove'")
            df.loc[df.udense > thr_udense, 'kind'] = 'background'
            df['name'] = name
            df['stage'] = j
            df['size'] = 1
            df['symbol'] = 1
            dfs.append(df)
            num.append(df.query("kind == 'cell'").shape[0])
    df = pd.concat(dfs, axis=0)
    fig = px.scatter(
        df,
        x='udense',
        y='signal',
        color='kind',
        size='size',
        size_max=3,
        opacity=0.5,
        facet_col='stage',
        facet_row='name',
    )
    for i in range(len(fig.data)):
        fig.data[i].marker.line.width = 0
    for _i, name in enumerate(paths.keys()):
        fig.update_annotation(
            selector={'text': f'naem={name}'},
            text=name,
        )
    for _j, _stage in enumerate(paths.keys()):
        fig.update_annotation(
            selector={'text': f'naem={name}'},
            text='',
        )
    fig.update_layout(
        template='none',
        font_size=11,
        showlegend=False,
        width=600,
        height=300,
        margin={'l': 35, 'r': 10, 't': 20, 'b': 32},
    )
    return fig



import hydra

from .cui import run
from .cui import test


@hydra.main(version_base=None, config_path=None, config_name='config')
def main(cfg):
    match cfg.mode:
        case 'test':
            test(cfg)
        case 'run':
            run(cfg)
        case _:
            raise ValueError()


if __name__ == '__main__':
    main()



from .clip import get_clip
from .gpu import delete_xla_buffers
from .gpu import from_tf
from .gpu import get_gpu_env
from .gpu import get_gpu_info
from .gpu import get_gpu_used
from .gpu import get_xla_stats
from .progress import get_progress
from .timer import Timer

__all__ = [
    'Data',
    'Timer',
    'get_clip',
    'get_progress',
    'from_tf',
    'get_gpu_env',
    'get_gpu_info',
    'get_gpu_used',
    'get_xla_stats',
    'delete_xla_buffers',
]



import numpy as np
import pandas as pd
import plotly.express as px
from PIL import Image
from PIL import ImageDraw

from ..cui.common import load
from ..spike import get_dynamics
from .common import to_image


def to_csv(cfg, stage, path, *, tsel=None, ksel=None):
    u, _, _, stats = load(cfg, 'temporal', stage)
    u = u[stats.query('kind == "cell"').spkid, :]

    dynamics = get_dynamics(cfg.dynamics)
    pad = dynamics.size - 1
    v = dynamics(u)
    u = u[:, pad:]

    if tsel is None:
        tsel = np.arange(u.shape[1])
    if ksel is None:
        ksel = np.arange(u.shape[0])

    udata = {'frame': tsel, 'time': tsel / cfg.data.imgs.hz}
    vdata = {'frame': tsel, 'time': tsel / cfg.data.imgs.hz}
    for k in ksel:
        udata[f'cell{k:04}'] = u[k, tsel]
        vdata[f'cell{k:04}'] = v[k, tsel]
    pd.DataFrame(udata).to_csv(path / 'spike.csv')
    pd.DataFrame(vdata).to_csv(path / 'calcium.csv')
    #_, _, _, stats = load(cfg, 'temporal', stage)
    #stats = stats.query("kind == 'cell'")
    ## u = u[stats.udense <= thr_udense]
    #return _spike_image(u, tsel, ksel, width, lines)


def spike_image(cfg, stage, tsel=slice(None), ksel=slice(None), width=3, lines=(), thr_udense=1.0):
    pad = get_dynamics(cfg.dynamics).size - 1
    u, _, _, stats = load(cfg, 'temporal', stage)
    u = u[stats.query('kind == "cell"').spkid, pad:]
    #_, _, _, stats = load(cfg, 'temporal', stage)
    #stats = stats.query("kind == 'cell'")
    # u = u[stats.udense <= thr_udense]
    return _spike_image(u, tsel, ksel, width, lines)


def _spike_image(u, tsel=slice(None), ksel=slice(None), width=3, lines=()):
    u /= u.max(axis=1, keepdims=True)
    u = u[ksel, tsel]
    nk, nt = u.shape
    img = to_image(u, 'Reds')
    img = Image.fromarray(img)
    draw = ImageDraw.Draw(img)
    draw.line(
        (int(0.8 * nt), (nk - 10), int(0.9 * nt), (nk - 10)),
        fill=(0, 0, 0),
        width=width,
    )
    for k0, k1, color, width in lines:
        if k0 < 0:
            k0 = nk + k0
        if k1 < 0:
            k1 = nk + k1
        draw.line((0, k0, 0, k1), fill=color, width=width)
    return img, nt, nk


def spike_stats_fig(cfg, stages, **kwargs):
    kwargs.setdefault('template', 'none')
    kwargs.setdefault('margin', {'l': 40, 'r': 10, 't': 20, 'b': 35})

    dfs = []
    for stage in stages:
        _, _, stats, _ = load(cfg, 'temporal', stage)
        stats['epoch'] = stage
        dfs.append(stats)
    stats = pd.concat(dfs, axis=0)
    fig = px.scatter(
        stats,
        x='signal',
        y='udense',
        opacity=0.3,
        facet_col='epoch',
        labels={'signal': 'signal intensity', 'udense': 'spike density'},
    )
    fig.update_layout(
        **kwargs,
    )
    return fig



from logging import getLogger

import numpy as np
import pandas as pd

from ..utils.math import calc_mah
from ..utils.math import calc_sn
from ..utils.math import robust_zscore
from .dynamics import get_dynamics
from .dynamics import get_rdynamics

logger = getLogger(__name__)


def evaluate(stats, spikes, bg):
    cond_cell = stats.kind == 'cell'
    cond_bg = stats.kind == 'background'
    cmax = spikes.max(axis=1)

    active_cell = cmax > 0
    cond_remove_cell = np.zeros_like(cond_cell)
    cond_remove_cell[cond_cell] = ~active_cell
    cond_active_cell = np.zeros_like(cond_cell)
    cond_active_cell[cond_cell] = active_cell

    spikes = spikes[active_cell]
    cmax = cmax[active_cell]

    csd = spikes.mean(axis=1) / cmax
    cnonzero = np.count_nonzero(spikes > 0, axis=1)
    csn = calc_sn(spikes)
    crsn = 1 / csn
    czrsn = robust_zscore(crsn)
    x = stats.loc[cond_active_cell, 'firmness'].to_numpy()
    y = crsn
    chypot = np.hypot(robust_zscore(x), robust_zscore(y))
    cmah = calc_mah(x, y)

    bmax = np.abs(bg).max(axis=1)
    bsn = np.array([bi.max() / (1.4826 * np.median(np.abs(bi[bi != 0]))) for bi in bg])

    stats.loc[cond_remove_cell, 'kind'] = 'remove'

    stats['spkid'] = -1
    stats.loc[cond_active_cell, 'spkid'] = np.arange(spikes.shape[0])
    stats['signal'] = None
    stats.loc[cond_active_cell, 'signal'] = cmax
    stats['udense'] = None
    stats.loc[cond_active_cell, 'udense'] = csd
    stats['unz'] = -1
    stats.loc[cond_active_cell, 'unz'] = cnonzero
    stats['snratio'] = None
    stats.loc[cond_active_cell, 'snratio'] = csn
    stats['rsn'] = None
    stats.loc[cond_active_cell, 'rsn'] = crsn
    stats['zrsn'] = None
    stats.loc[cond_active_cell, 'zrsn'] = czrsn
    stats['hypot'] = None
    stats.loc[cond_active_cell, 'hypot'] = chypot
    stats['mah'] = None
    stats.loc[cond_active_cell, 'mah'] = cmah

    stats['bgid'] = -1
    stats.loc[cond_bg, 'bgid'] = np.arange(bg.shape[0])
    stats['bmax'] = None
    stats.loc[cond_bg, 'bmax'] = bmax
    stats['bsparse'] = None
    stats.loc[cond_bg, 'bsparse'] = bsn

    labels = [
        'kind',
        'old_kind',
        'segid',
        'spkid',
        'bgid',
        'y',
        'x',
        'radius',
        'aratio',
        'firmness',
        'pos_move',
        'min_dist_id',
        'min_dist',
        'max_dup_id',
        'max_dup',
        'asum',
        'area',
        'signal',
        'udense',
        'unz',
        'rsn',
        'zrsn',
        'hypot',
        'mah',
        'bmax',
        'bsparse',
    ]
    return stats[[k for k in labels if k in stats.columns]], spikes, bg


def fix_kind(stats, spikes, bg, dynamics, bg_type='bg', thr_bg=None, thr_cell=None):
    if thr_bg is None:
        thr_bg = {}
    if thr_cell is None:
        thr_cell = {}

    fdynamics = get_dynamics(dynamics)
    rdynamics = get_rdynamics(dynamics)
    cell_df = stats.query("kind == 'cell'")
    spikes = spikes[cell_df.spkid.to_numpy()]
    bg_df = stats.query("kind == 'background'")
    logger.info('thr %s %s', thr_bg, thr_cell)
    with np.printoptions(precision=3, suppress=True):
        bins = [-np.inf, -4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0, 5.0, np.inf]
        hist, bins = np.histogram(cell_df.zrsn, bins=bins)
        for s, e, c in zip(bins[:-1], bins[1:], hist, strict=False):
            logger.info('[%f %f): %d', s, e, c)
        bins = [0.0, 0.2, 0.3, 0.35, 0.4, 0.45, 0.5, 0.6, np.inf]
        hist, bins = np.histogram(cell_df.firmness, bins=bins)
        for s, e, c in zip(bins[:-1], bins[1:], hist, strict=False):
            logger.info('[%f %f): %d', s, e, c)

    if bg_type == 'remove':
        remove_mask = np.zeros(cell_df.shape[0], bool)
        for k, thr in thr_bg.items():
            if k in ['rsn', 'zrsn', 'hypot', 'mah']:
                remove_mask |= cell_df[k] > thr
            else:
                remove_mask |= cell_df[k] < thr
        spikes = spikes[~remove_mask]
        cell_df = cell_df.loc[~remove_mask].copy()
        cell_df['spkid'] = np.arange(cell_df.shape[0])

        stats = cell_df
    elif bg_type == 'bg':
        to_bg_mask = np.zeros(cell_df.shape[0], bool)
        for k, thr in thr_bg.items():
            if k in ['rsn', 'zrsn', 'hypot', 'mah']:
                to_bg_mask |= cell_df[k] > thr
            else:
                to_bg_mask |= cell_df[k] < thr

        if len(thr_cell) == 0:
            to_cell_mask = np.zeros(bg_df.shape[0], bool)
        else:
            to_cell_mask = np.ones(bg_df.shape[0], bool)
            for k, thr in thr_cell.items():
                to_cell_mask &= bg_df[k] > thr

        spikes, to_bg = spikes[~to_bg_mask], spikes[to_bg_mask]
        cell_df, to_bg_df = (
            cell_df.loc[~to_bg_mask].copy(),
            cell_df.loc[to_bg_mask].copy(),
        )
        bg, to_cell = bg[~to_cell_mask], bg[to_cell_mask]
        bg_df, to_cell_df = (
            bg_df.loc[~to_cell_mask].copy(),
            bg_df.loc[to_cell_mask].copy(),
        )
        to_bg = np.array(fdynamics(to_bg))
        to_cell = np.array(rdynamics(to_cell))

        spikes = np.concatenate([spikes, to_cell], axis=0)
        bg = np.concatenate([bg, to_bg], axis=0)

        nc = cell_df.shape[0]
        nb = bg_df.shape[0]
        nn = to_bg_df.shape[0]
        nm = to_cell_df.shape[0]

        cell_df['spkid'] = np.arange(cell_df.shape[0])
        bg_df['bgid'] = np.arange(bg_df.shape[0])

        to_bg_df['kind'] = 'background'
        to_bg_df['spkid'] = -1
        to_bg_df['bgid'] = np.arange(nb, nb + nn, dtype=np.int32)

        to_cell_df['kind'] = 'cell'
        to_cell_df['spkid'] = np.arange(nc, nc + nm, dtype=np.int32)
        to_cell_df['bgid'] = -1

        stats = pd.concat([cell_df, to_cell_df, bg_df, to_bg_df], axis=0)
    else:
        raise ValueError()
    return stats, spikes, bg



from .peaks import peak_stats_fig
from .seg import seg_max_fig
from .seg import to_multipage_tif
from .spike import spike_image
from .spike import to_csv

__all__ = [
    'peak_stats_fig',
    'seg_max_fig',
    'spike_iamge',
    'to_csv',
    'to_multipage_tif',
]



from logging import getLogger

import numpy as np

from ..footprint import find_peaks
from ..footprint import make_footprints
from ..footprint import reduce_peaks
from ..io import save
from ..io import try_load
from ..utils import get_xla_stats
from .common import get_data
from .common import get_files
from .common import get_force
from .common import reduce_log

logger = getLogger(__name__)


def init(cfg):
    (statsfile,) = get_files(cfg, 'init', 0)
    if get_force(cfg, 'init', 0) or not statsfile.exists():
        (makefile,) = get_files(cfg, 'make', 0)
        (reducefile,) = get_files(cfg, 'reduce', 0)
        if get_force(cfg, 'make', 0) or not makefile.exists():
            logger.debug('%s', get_xla_stats())
            data = get_data(cfg)
            if get_force(cfg, 'reduce', 0) or not reducefile.exists():
                (findfile,) = get_files(cfg, 'find', 0)
                if get_force(cfg, 'find', 0) or not findfile.exists():
                    logger.info('exec find')
                    findval = find_peaks(data, cfg.radius.filter, **cfg.cmd.find)
                    logger.debug('%s', get_xla_stats())
                    save(findfile, findval)
                    logger.info('saved find')
                else:
                    logger.info('load find')
                    findval = try_load(findfile)

                logger.info('exec reduce')
                peaks = reduce_peaks(
                    findval,
                    bg_type=cfg.bg_type,
                    **cfg.init.args,
                    **cfg.cmd.reduce,
                )
                logger.debug('%s', get_xla_stats())
                save(reducefile, peaks)
                if cfg.cmd.remove_find:
                    findfile.unlink(missing_ok=True)
                logger.info('saved reduce')
            else:
                logger.info('load reduce')
                peaks = try_load(reducefile)
            reduce_log(cfg, 0)

            logger.info('exec make')
            footprints = make_footprints(data, peaks, **cfg.cmd.make)
            logger.debug('%s', get_xla_stats())
            save(makefile, footprints)
            logger.info('saved make')
        else:
            logger.info('load make')
            footprints = try_load(makefile)
            peaks = try_load(reducefile)
        logger.info('exec init')
        peaks['asum'] = footprints.sum(axis=(1, 2))
        peaks['area'] = np.count_nonzero(footprints > 0, axis=(1, 2))
        peaks.loc[peaks['asum'] == 0, 'kind'] = 'remove'
        save(statsfile, peaks)
        if cfg.cmd.remove_reduce:
            reducefile.unlink(missing_ok=True)
        logger.info('saved init')



from logging import getLogger
from pathlib import Path

import numpy as np
from tifffile import TiffFile
from tifffile import imread
from tifffile import memmap

logger = getLogger(__name__)


def load_imgs(**cfg):
    logger.debug('load_imgs: %s', cfg)
    match cfg['type']:
        case 'npy':
            imgs = np.load(cfg['file'], mmap_mode='r')
        case 'raw':
            dtype = np.dtype(cfg['dtype']).newbyteorder(cfg['endian'])
            data = np.memmap(cfg['file'], dtype, 'r')
            imgs = data.reshape(-1, cfg['height'], cfg['width'])
        case 'tif':
            imgsfile = Path(cfg['file'])
            imgsfile_fix = imgsfile.with_stem(f'{imgsfile.stem}_fix')
            if imgsfile_fix.exists():
                imgsfile = imgsfile_fix
            with TiffFile(imgsfile) as tif:
                data = tif.series[0]
                if data.dataoffset is None:
                    logger.debug('create fixed tiff: %s', imgsfile_fix)
                    imgs = memmap(imgsfile_fix, shape=data.shape, dtype=data.dtype)
                    for i, pi in enumerate(data):
                        imgs[i] = pi.asarray()
                else:
                    imgs = data.asarray(out='memmap')
            logger.debug('mount tiff as memmap: %s', imgsfile_fix)
        case _:
            raise RuntimeError(f'{cfg["file"]} is not imgs file')
    return imgs, cfg['hz']


def apply_mask(imgs, **cfg):
    logger.debug('apply_mask: %s', cfg)
    match cfg['type']:
        case 'nomask':
            mask = None
        case 'tif':
            mask = imread(cfg['file']) > 0
        case 'npy':
            mask = np.load(cfg['file']) > 0
        case _:
            raise RuntimeError('bad file type: {maskfile}')

    if mask is not None:
        my = np.where(np.any(mask, axis=1))[0]
        mx = np.where(np.any(mask, axis=0))[0]
        x0, y0, w, h = mx[0], my[0], mx[-1] - mx[0] + 1, my[-1] - my[0] + 1
        imgs = imgs[:, y0 : y0 + h, x0 : x0 + w]
        mask = mask[y0 : y0 + h, x0 : x0 + w]
        logger.debug('data size: (%d, %d, %d)', *imgs.shape)

    nt, h, w = imgs.shape
    nx = -1 if mask is None else h * w
    logger.info('imgs, mask: (%d, %d, %d), %d', nt, h, w, nx)
    return imgs, mask



from logging import getLogger

from ..filter import movie_stats
from ..io import apply_mask
from ..io import load_imgs
from ..io import save
from ..utils import get_xla_stats
from .common import get_files
from .common import get_force

logger = getLogger(__name__)


def normalize(cfg):
    force = get_force(cfg, 'normalize', 0)
    files = get_files(cfg, 'normalize', 0)
    if force or not all(file.exists() for file in files):
        logger.info('exec normalize')
        imgs, hz = load_imgs(**cfg.data.imgs)
        imgs, mask = apply_mask(imgs, **cfg.data.mask)
        logger.info('%s', get_xla_stats())
        outputs = movie_stats(imgs, mask, **cfg.cmd.stats)
        logger.info('%s', get_xla_stats())
        save(files, outputs)
        logger.info('saved normalize')



from .laplace import gaussian_laplace
from .laplace import gaussian_laplace_single
from .map import mapped_imgs
from .pool import max_pool
from .stats import movie_stats

__all__ = [
    'mapped_imgs',
    'max_pool',
    'gaussian_laplace',
    'gaussian_laplace_single',
    'movie_stats',
]



from .model import SpatialModel
from .model import TemporalModel
from .penalty import Penalty
from .penalty import get_penalty
from .regularizer import MaxNormNonNegativeL1
from .regularizer import NonNegativeL1
from .regularizer import Regularizer

__all__ = [
    'Regularizer',
    'NonNegativeL1',
    'MaxNormNonNegativeL1',
    'Penalty',
    'SpikeToCalcium',
    'get_dynamics',
    'get_penalty',
    'SpatialModel',
    'TemporalModel',
]



from logging import getLogger

import numpy as np
import pandas as pd

from ..io import save
from ..io import try_load
from ..spike import evaluate
from ..train import TemporalModel
from ..train import get_penalty
from ..utils import get_clip
from ..utils import get_xla_stats
from .common import get_data
from .common import get_files
from .common import get_force
from .common import rev_index

logger = getLogger(__name__)


def temporal(cfg, stage, *, force=False):
    spikefile, bgfile, lossfile, statsfile = get_files(cfg, 'temporal', stage)
    if (
        get_force(cfg, 'temporal', stage)
        or not spikefile.exists()
        or not bgfile.exists()
        or not statsfile.exists()
    ):
        if stage == 0:
            footprints = try_load(get_files(cfg, 'make', stage))
            stats = try_load(get_files(cfg, 'init', stage))
        else:
            stats, footprints = try_load(get_files(cfg, 'clean', stage))
        logger.info(f'exec temporal ({stage})')
        data = get_data(cfg)
        logger.debug('%s', get_xla_stats())
        model = TemporalModel(
            data,
            stats,
            footprints,
            cfg.dynamics,
            get_penalty(cfg.penalty, stage),
        )
        clips = get_clip(data.shape, cfg.cmd.temporal.clip)
        logdfs = []
        out = []
        for i, clip in enumerate(clips):
            model.prepare(clip, **cfg.cmd.temporal.prepare)
            log = model.fit(**cfg.cmd.temporal.step)
            logger.debug('%s', get_xla_stats())

            loss, sigma = zip(*log, strict=False)
            df = pd.DataFrame(
                {
                    'stage': stage,
                    'kind': 'temporal',
                    'div': i,
                    'step': np.arange(len(log)),
                    'loss': loss,
                    'sigma': sigma,
                },
            )
            logdfs.append(df)
            out.append(model.get_x())
        logdf = pd.concat(logdfs, axis=0)
        index1, index2, x1, x2 = (np.concatenate(v, axis=0) for v in zip(*out, strict=False))
        spikes = np.array(x1[rev_index(index1)])
        bg = np.array(x2[rev_index(index2)])
        if cfg.fix_top:
            nk, nu = spikes.shape
            idx = np.argpartition(spikes, -2, axis=1)
            spikes[np.arange(nk), idx[:, -1]] = spikes[np.arange(nk), idx[:, -2]]
        stats, spikes, bg = evaluate(stats, spikes, bg)
        save((spikefile, bgfile, lossfile, statsfile), (spikes, bg, logdf, stats))
        logger.info(f'saved temporal ({stage})')



from collections import namedtuple
from logging import getLogger

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
import tensorflow as tf
from scipy.ndimage import grey_closing

from ..filter import gaussian_laplace
from ..utils import from_tf
from ..utils import get_gpu_env
from ..utils.math import calc_mah
from ..utils.math import robust_zscore
from .radius import get_radius
from .segment import get_segment_mask

# from sklearn.mixture import GaussianMixture as CluModel


logger = getLogger(__name__)

Footprint = namedtuple('Footprint', 'foootprit y x radius intensity')


def clean(
    stats,
    segs,
    radius,
    bg_type='bg',
    cell_range=None,
    thr_remove=None,
    thr_bg=None,
    thr_cell=None,
    env=None,
    factor=1,
    prefetch=1,
):
    segments, y, x, radius, firmness = clean_footprints(
        segs,
        radius,
        env,
        factor,
        prefetch,
    )

    ratio = []
    for ai in segments:
        gy, gx = np.where(ai > 0)
        w = ai[gy, gx]
        if w.size > 0:
            try:
                s = np.sqrt(np.linalg.eigvalsh(np.cov(gy, gx, aweights=w, ddof=0)))
                v = s.max() / s.min()
            except np.linalg.LinAlgError:
                v = np.nan
        else:
            v = np.nan
        ratio.append(v)
    ratio = np.array(ratio, np.float32)

    stats = stats.sort_values('segid')
    oldy = stats.y.to_numpy()
    oldx = stats.x.to_numpy()
    oldr = stats.radius.to_numpy()

    stats['old_kind'] = stats.kind
    stats['y'] = y
    stats['x'] = x
    stats['radius'] = radius
    stats['aratio'] = ratio
    stats['firmness'] = firmness
    stats['pos_move'] = np.hypot(x - oldx, y - oldy) / oldr

    cell_df = stats.query("old_kind == 'cell'")
    firmness = cell_df.firmness.to_numpy()
    rsn = cell_df.rsn.to_numpy()
    hypot = np.hypot(robust_zscore(firmness), robust_zscore(rsn))
    mah = calc_mah(firmness, rsn)
    stats['hypot'] = None
    stats.loc[cell_df.index, 'hypot'] = hypot
    stats['mah'] = None
    stats.loc[cell_df.index, 'mah'] = mah

    bg_cond = stats.old_kind == 'background'
    to_cell_cond = bg_cond.copy()
    for k, thr in thr_cell.items():
        to_cell_cond &= stats[k] > thr
    bg_cond &= ~to_cell_cond

    remove_cond = np.zeros(stats.shape[0], bool)
    for k, thr in thr_remove.items():
        if k == 'pos_move':
            remove_cond |= (stats.old_kind == 'cell') & (stats[k] > thr)
        elif k == 'radius':
            remove_cond |= stats[k] < thr
        else:
            raise ValueError()

    tmp_cond = np.zeros(stats.shape[0], bool)
    for k, thr in thr_bg.items():
        tmp_cond |= stats[k] > thr
    if bg_type == 'remove':
        remove_cond |= tmp_cond
    elif bg_type == 'bg':
        bg_cond |= tmp_cond
    else:
        raise ValueError()

    bg_cond = bg_cond.to_numpy()
    remove_cond = remove_cond.to_numpy()

    stats['kind'] = 'cell'
    stats.loc[bg_cond, 'kind'] = 'background'
    stats.loc[remove_cond, 'kind'] = 'remove'
    with pd.option_context('display.max_rows', None):
        print(stats[['kind', 'old_kind', 'segid', 'radius', 'pos_move', 'aratio']])

    cell_cond = (stats.kind == 'cell').to_numpy()
    ys = y[cell_cond]
    xs = x[cell_cond]
    rs = radius[cell_cond]

    dist_mat = np.hypot(xs[:, np.newaxis] - xs, ys[:, np.newaxis] - ys) / rs
    np.fill_diagonal(dist_mat, np.inf)

    val_mat = segments[cell_cond][:, ys, xs]
    np.fill_diagonal(val_mat, 0)

    stats['max_dup_id'] = -1
    stats['min_dist_id'] = -1
    cell_df = stats.loc[cell_cond].copy()
    cell_df['min_dist_id'] = cell_df.index[np.argmin(dist_mat, axis=0)]
    cell_df['min_dist'] = dist_mat.min(axis=0)
    cell_df['max_dup_id'] = cell_df.index[np.argmax(val_mat, axis=0)]
    cell_df['max_dup'] = val_mat.max(axis=0)
    bg_df = stats.query("kind == 'background'").copy()
    remove_df = stats.query("kind == 'remove'").copy()
    stats = pd.concat([cell_df, bg_df, remove_df], axis=0)

    stats['spkid'] = -1
    stats['bgid'] = -1

    nc = cell_df.shape[0]
    nb = bg_df.shape[0]
    nr = remove_df.shape[0]
    logger.info('cell/bg/remove = %d/%d/%d', nc, nb, nr)
    return stats, segments


def clean_footprints(segs, radius, env=None, factor=1, prefetch=1):
    @jax.jit
    def calc(imgs):
        nk, h, w = imgs.shape
        k = jnp.arange(nk)
        imgs /= imgs.max(axis=(1, 2), keepdims=True)
        gl = gaussian_laplace(imgs, radius, -3)
        nk, nr, h, w = gl.shape
        idx = jnp.argmax(gl.reshape(nk, nr * h * w), axis=1)
        r, y, x = idx // (h * w), (idx // w) % h, idx % w
        g = gl[k, r, y, x]
        seg = jax.vmap(get_segment_mask)(gl[k, r], y, x)
        imgs = jnp.where(seg, imgs, 0)
        dmin = jnp.min(jnp.where(seg, imgs, jnp.inf), axis=(1, 2), keepdims=True)
        dmax = jnp.max(jnp.where(seg, imgs, -jnp.inf), axis=(1, 2), keepdims=True)
        imgs = jnp.where(seg, (imgs - dmin) / (dmax - dmin), 0)
        return imgs, y, x, r, g

    radius = get_radius(radius)
    nk, h, w = segs.shape

    env = get_gpu_env(env)
    nd = env.num_devices
    sharding = env.sharding((nd, 1))
    batch = env.batch(float(factor) * h * w * len(radius), nk)

    dataset = tf.data.Dataset.from_generator(
        lambda: zip(range(nk), segs, strict=False),
        output_signature=(
            tf.TensorSpec((), tf.int32),
            tf.TensorSpec((h, w), tf.float32),
        ),
    )
    dataset = dataset.batch(batch)
    dataset = dataset.prefetch(prefetch)

    out = jnp.empty((nk + 1, h, w))
    y = jnp.empty((nk + 1,), jnp.int32)
    x = jnp.empty((nk + 1,), jnp.int32)
    r = jnp.empty((nk + 1,), jnp.int32)
    g = jnp.empty((nk + 1,), jnp.float32)

    logger.debug('clean: %s %s', (factor, h, w), batch)
    logger.info('%s: %s %s %d', 'pbar', 'start', 'clean', nk)
    for data in dataset:
        data = (from_tf(v) for v in data)
        idx, img = (jax.device_put(v, sharding) for v in data)

        count = idx.size
        diff = batch - count
        if diff > 0:
            pad = (0, diff), (0, 0), (0, 0)
            idx = jnp.pad(idx, pad[:1], constant_values=-1)
            img = jnp.pad(img, pad, constant_values=jnp.nan)

        outi, yi, xi, ri, gi = calc(img)
        out = out.at[idx].set(outi)
        y = y.at[idx].set(yi)
        x = x.at[idx].set(xi)
        r = r.at[idx].set(ri)
        g = g.at[idx].set(gi)
        logger.info('%s: %s %d', 'pbar', 'update', count)
    logger.info('%s: %s', 'pbar', 'close')

    out, y, x, r, g = (np.array(v[:-1]) for v in (out, y, x, r, g))
    out = grey_closing(out, (1, 10, 10))
    r = np.array(radius)[r]
    return Footprint(out, y, x, r, g)


def get_dupfilter(**args):
    match args:
        case {'thr_active_area': thr}:
            return DupFilter(thr)
        case _:
            raise ValueError()


def get_bgfilter(**args):
    match args:
        case {'kind': 'new', **kwargs}:
            return BackgroundFilter(**kwargs)
        case {'kind': 'simple', **kwargs}:
            return SimpleBackgroundFilter(**kwargs)
        case {'kind': 'factor', **kwargs}:
            return FactorBackgroundFilter(**kwargs)
        case _:
            return OldBackgroundFilter(**args)


class DupFilter:
    def __init__(self, thr_active_area):
        self._thr = thr_active_area

    def set(self, stats, segments):
        self._y = stats.y.to_numpy()
        self._x = stats.x.to_numpy()
        self._segmask = segments > self._thr

    def dup_id(self, i, js):
        yi, xi = self._y[i], self._x[i]
        for j in js:
            yj, xj = self._y[j], self._x[j]
            if self._segmask[i, yj, xj] and self._segmask[j, yi, xi]:
                return j
        return -1


class BackgroundFilter:
    def __init__(
        self,
        bias=0,
        coef_log_snr=0,
        coef_firmness=0,
    ):
        self._bias = bias
        self._coef_log_snr = coef_log_snr
        self._coef_firmness = coef_firmness

    def set(self, stats):
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        score = (
            self._bias
            + self._coef_log_snr * np.log10(si.snratio)
            + self._coef_firmness * si.firmness
        )
        return score < 0


class FactorBackgroundFilter:
    def __init__(
        self,
        thr_rsn_factor=10,
        thr_firmness=0,
    ):
        self._thr_s = thr_rsn_factor
        self._thr_f = thr_firmness

    def set(self, stats):
        stats['rsn'] = 1 / stats.snratio
        med = np.median(stats.rsn)
        std = 1.4826 * np.median(np.abs(stats.rsn - med))
        stats['zrsn'] = (stats.rsn - med) / std
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        return (si.zrsn > self._thr_s) or (si.firmness < self._thr_f)


class SimpleBackgroundFilter:
    def __init__(
        self,
        thr_snr_inv=1,
        thr_firmness=0,
    ):
        self._thr_s = thr_snr_inv
        self._thr_f = thr_firmness

    def set(self, stats):
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.iloc[i]
        return (1 / si.snratio > self._thr_s) or (si.firmness < self._thr_f)


class OldBackgroundFilter:
    def __init__(
        self,
        thr_bg_udense=1,
        thr_bg_signal=0,
        thr_bg_firmness=0,
        thr_bg_cluster=1,
    ):
        self._thr_d = thr_bg_udense
        self._thr_s = thr_bg_signal
        self._thr_f = thr_bg_firmness
        self._thr_c = thr_bg_cluster

    def set(self, stats):
        """
        cindex = oldstats.query("kind == 'cell'").cid
        v = np.stack([udense, firmness, np.log(signal)], axis=1)
        v = v[cindex]
        clu = CluModel(
            n_components=2,
            weights_init=[0.95, 0.05],
            means_init=[[0.03, 0.45, 0.5], [0.13, 0.35, -0.3]],
        )
        clu.fit(v)
        z[cindex] = clu.predict_proba(v)[:, 1]
        self._stats.
        """
        self._stats = stats

    def is_background(self, i):
        s = self._stats
        si = s.at[s.index[i]]
        return (
            # (si.z > thr_bg_cluster)
            (si.udense > self._thr_d) or (si.signal < self._thr_s) or (si.firmness < self._thr_f)
        )



from .image import apply_mask
from .image import load_imgs
from .movie import gen_result_movie
from .saver import save
from .saver import try_load

__all__ = [
    'load_imgs',
    'apply_mask',
    'save',
    'try_load',
    'gen_result_movie',
]



import pandas as pd
import plotly.graph_objects as go

from .common import add_jitter


def compare_lu_fig(paths, labels):
    fig = go.Figure().set_subplots(
        2,
        len(paths),
        row_heights=(1, 5),
        horizontal_spacing=0.05,
        vertical_spacing=0.01,
        # shared_xaxes="columns",
        # shared_yaxes="rows",
    )
    for i, path in enumerate(paths):
        stats = pd.read_csv(path, index_col=0)
        df = stats.query("kind != 'remove'").copy()
        df = add_jitter(df)
        fig.add_trace(
            go.Histogram(x=df.udense, marker_color='grey'),
            row=1,
            col=i + 1,
        )
        fig.add_trace(
            go.Scatter(
                y=df.intensity,
                x=df.udense,
                mode='markers',
                marker={'size': 3, 'color': 'blue', 'opacity': 0.5},
            ),
            row=2,
            col=i + 1,
        )
        fig.add_annotation(
            x=0.2,
            y=15,
            text=labels[i],
            showarrow=False,
            font_size=11,
            row=2,
            col=i + 1,
        )
        fig.update_xaxes(
            showticklabels=False,
            row=1,
            col=i + 1,
        )
        fig.update_yaxes(
            showticklabels=False,
            row=1,
            col=i + 1,
        )
        fig.update_xaxes(
            range=(0, 0.28),
            tickmode='array',
            tickvals=[0, 0.1, 0.2],
            ticktext=[0, 0.1, 0.2],
            title_text='spike density',
            row=2,
            col=i + 1,
        )
        fig.update_yaxes(
            title_text='intensity' if i == 0 else '',
            showticklabels=i == 0,
            # range=((4 / 5), (21 / 5)),
            # tickmode="array",
            # tickvals=np.log2([3, 6, 12]),
            # ticktext=["3", "6", "12"],
            row=2,
            col=i + 1,
        )
    fig.update_layout(
        template='none',
        font_size=11,
        width=600,
        height=200,
        showlegend=False,
        margin={'t': 0, 'r': 10, 'l': 40, 'b': 40},
    )
    return fig



from logging import getLogger

import av
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont

logger = getLogger(__name__)


def gen_movie(filename, imgs, shape, fps, **kwargs):
    logger.info('%s: %s %s %d', 'pbar', 'start', 'movie', shape[0])
    with av.open(filename, 'w') as output:
        # codec = av.CodecContext.create("h264_nvenc", "w")
        # codec = av.CodecContext.create("libx264", "w")
        stream = output.add_stream(kwargs.get('codec', 'h264'), fps)
        # stream = output.add_stream(codec, fps)
        stream.bit_rate = kwargs.get('bit_rate', kwargs.get('bit_rate', 8_000_000))
        stream.pix_fmt = kwargs.get('fmt', kwargs.get('fmt', 'yuv420p'))
        stream.height = shape[1]
        stream.width = shape[2]
        for img in imgs:
            frame = av.VideoFrame.from_ndarray(img, format='rgba')
            packet = stream.encode(frame)
            output.mux(packet)
            logger.info('%s: %s %d', 'pbar', 'update', 1)
        logger.info('%s: %s', 'pbar', 'close')


def gen_normalize_movie(outfile, data, **kwargs):
    imgs, mask, hz, avgx, avgt, std0, min0, max0, min1, max1 = data
    nt, h, w = imgs.shape

    def gen_frame():
        for t in range(nt):
            img0 = (imgs[t] - min0) / (max0 - min0)
            img1 = (imgs[t] - avgx - avgt[t]) / std0
            img1 = (img1 - min1) / (max1 - min1)
            img = np.concatenate([img0, img1], axis=1)
            img = (255 * plt.get_cmap('Greens')(img)).astype(np.uint8)
            if mask is not None:
                img[~mask] = 0
            yield (255 * img).astype(np.uint8)

    gen_movie(outfile, gen_frame(), (nt, h, 2 * w), hz, **kwargs)


def gen_result_movie(outfile, data, footprints, spikes, dynamics, t0=None, t1=None, **kwargs):
    calcium = np.array(dynamics(spikes))
    # calcium -= calcium.min(axis=1, keepdims=True)
    # calcium -= np.median(calcium, axis=1, keepdims=True)
    # calcium = np.where(calcium > 0, calcium, 0.0)
    calcium /= calcium.max()
    if t0 is None:
        t0 = 0
    if t1 is None:
        t1 = calcium.shape[1]
    cmap = plt.get_cmap('Greens')

    dmin = jnp.array(data.min1)
    dmax = jnp.array(data.max1)
    calcium = jnp.array(calcium)
    footprints = jnp.array(footprints)

    def gen(data, t):
        img0 = (data - dmin) / (dmax - dmin)
        img1 = (calcium[:, t, jnp.newaxis, jnp.newaxis] * footprints).sum(axis=0)
        img1 = jnp.where(img1 < 1.0, img1, 1.0)

        img = np.concatenate([img0, img1], axis=1)
        img = (255 * cmap(img)).astype(np.uint8)
        return img

    def gen_frame():
        for t in range(t0, t1):
            img = np.array(gen(jnp.array(data.select(t)), jnp.array(t)))
            img = Image.fromarray(img)
            draw = ImageDraw.Draw(img)
            draw.text((10, 10), 'Normalized original movie', 'red', font=font)
            draw.text((w + 10, 10), 'Reconstructed movie', 'red', font=font)

            yield np.array(img)

    font = ImageFont.truetype('/usr/share/fonts/truetype/ttf-bitstream-vera/VeraBd.ttf', 24)
    nk, _ = calcium.shape
    _, h, w = footprints.shape
    footprints = footprints[:nk]
    gen_movie(outfile, gen_frame(), (t1 - t0, h, 2 * w), data.hz, **kwargs)



import os
from logging import getLogger
from pathlib import Path

import numpy as np
import pandas as pd
from hydra.core.hydra_config import HydraConfig

from ..io import apply_mask
from ..io import load_imgs
from ..io import try_load
from .data import Data

logger = getLogger(__name__)


def set_env(cfg):
    hcfg = HydraConfig.get()
    for k, v in hcfg.job.env_set.items():
        os.environ[k] = v


def get_force(cfg, name, stage):
    force_dict = {
        'normalize': [
            'normalize',
            'find',
            'reduce',
            'make',
            'init',
            'temporal',
        ],
        'find': ['find', 'reduce', 'make', 'init', 'temporal'],
        'reduce': ['reduce', 'make', 'init', 'temporal'],
        'make': ['make', 'init', 'temporal'],
        'init': ['init', 'temporal'],
        'spatial': ['spatial', 'clean', 'temporal'],
        'clean': ['clean', 'temporal'],
        'temporal': ['temporal'],
    }
    force = (stage > cfg.force_from[0]) or (
        (stage == cfg.force_from[0]) and (name in force_dict[cfg.force_from[1]])
    )
    logger.debug('test force: %s %d', name, stage)
    return force


def all_stats(cfg):
    out = []
    for stage in range(1000):
        (_, _, _, stats), flag = load(cfg, 'temporal', stage)
        if flag is None:
            break
        out.append(stats)
    return out


def get_files(cfg, name, stage):
    odir = Path(cfg.outputs.dir)
    path = cfg.outputs[name]
    fdir = odir / path.dir0 if stage == 0 or stage == 1 and name == 'spatial' else odir / path.dir
    files = [fdir / file.format(stage=stage) for file in path.files]
    return files


def load(cfg, name, stage):
    return try_load(get_files(cfg, name, stage))


def get_data(cfg):
    imgs, hz = load_imgs(**cfg.data.imgs)
    imgs, mask = apply_mask(imgs, **cfg.data.mask)
    stats = try_load(get_files(cfg, 'normalize', 0)[0])
    data = Data(imgs, mask, hz, *stats)
    return data


def rev_index(index):
    rev_index = np.full_like(index, index.size)
    for i, j in enumerate(index):
        rev_index[j] = i
    return rev_index


def reduce_log(cfg, stage):
    if stage == 0:
        stats = try_load(get_files(cfg, 'reduce', stage))
    else:
        stats, _ = try_load(get_files(cfg, 'clean', stage))
    cell = stats[stats.kind == 'cell']
    bg = stats[stats.kind == 'background']
    removed = stats[stats.kind == 'remove']

    radius = np.sort(np.unique(stats.radius))
    for r in radius:
        logger.info(
            'radius=%f cell=%d bg=%d remove=%d',
            r,
            (cell.radius == r).sum(),
            (bg.radius == r).sum(),
            (removed.radius == r).sum(),
        )


def print_stats(cfg, stage):
    _, _, _, stats = try_load(get_files(cfg, 'temporal', stage))

    cell = stats[stats.kind == 'cell']
    rsn = cell.rsn
    med = np.median(rsn)
    std = 1.4826 * np.median(np.abs(rsn - med))
    thr = med + 2 * std
    logger.info('rsn: med=%.3f std=%.3f thr=%.3f', med, std, thr)
    labels = [
        'y',
        'x',
        'radius',
        'firmness',
        'signal',
        # "udense",
        'unz',
        'rsn',
        'zrsn',
    ]
    if stage > 0:
        labels += [
            'min_dist_id',
            'min_dist',
            'max_dup_id',
            'max_dup',
        ]
    with pd.option_context('display.max_rows', None):
        logger.info(
            'cell: %d\n%s',
            cell.shape[0],
            cell[labels].sort_values('signal', ascending=False),
        )

    bg = stats[stats.kind == 'background']
    if bg.shape[0] > 0:
        labels = [
            'y',
            'x',
            'radius',
            'firmness',
            'bmax',
            'bsparse',
        ]
        # if stage > 0:
        #    labels += ["old_radius", "old_udense", "old_bsparse"]
        with pd.option_context('display.max_rows', None):
            logger.info(
                'background: %d\n%s',
                bg.shape[0],
                bg[labels].sort_values('bmax', ascending=False),
            )

    removed = stats[stats.kind == 'remove']
    if removed.shape[0] == 0:
        logger.info('removed: 0')
    else:
        labels = [
            'y',
            'x',
            'radius',
            'firmness',
        ]
        if stage > 0:
            labels += [
                'pos_move',
            ]
        with pd.option_context('display.max_rows', None):
            logger.info(
                'removed: %d\n%s',
                removed.shape[0],
                removed[labels].sort_values('firmness', ascending=False),
            )


def finish(cfg, stage):
    _, _, _, stats = try_load(get_files(cfg, 'temporal', stage))
    removed = stats[stats.kind == 'remove']

    # reduce_log(cfg, stage)
    return (stage > 0) and (removed.shape[0] <= cfg.early_stop)
